package algo.branch;

import algo.datastructure.*;
import algo.util.QuickSelect;

public class LowerBound {

    static int resetCounter = 0;
    static boolean flagUnfeasible = false;
    static boolean reset() {
        //in a fixed cycle the packing is computed from scratch
        return resetCounter % 7 == 0;
    }

    static int lb1Count = 0;
    static int lb2Count = 0;

    public static int lb1(Graph g, int t) {
        int[] ncArraySelected = new int[g.ncCount()];
        for (int i = 0; i < g.ncCount(); i++) ncArraySelected[i] = g.getNC(i).size;
        QuickSelect.selectSortTopKMax(ncArraySelected, t);

        int lb1 = 0;

        int maxSizeOfRest = 0;
        for (int i = t; i < ncArraySelected.length; i++) {
            int size = ncArraySelected[i];
            lb1 += size;
            maxSizeOfRest = Math.max(maxSizeOfRest, size);
        }

        return (lb1 + 1) / 2;
    }

    public static int lb2(Graph g, int t, int k, boolean useWeightedVersion) {
        int lb = 0;
        if (Packing.updateEverything || reset()) {//reset packing
            resetCounter++;
            Packing.updateEverything = false;
            Packing.initializePacking(g, t, useWeightedVersion);
        } else {//compute cost of current packing
            resetCounter++;
            for (Pack pack : Packing.packs) {
                if (pack.updatePackCount) {
                    pack.updatePackCount = false;
                    pack.computeCost(g);
                }
                if (pack.cost == Integer.MAX_VALUE) return k + 1;
                lb += pack.cost;
            }
            if ((lb + 1) / 2 > k) return k + 1;
        }
        //improve locally
        int increase;
        while ((increase = Packing.improvePackingBySwaps()) != 0) {
            if (flagUnfeasible) {
                flagUnfeasible = false;
                return k + 1;
            }
            lb += increase;
        }
        //search for new packs
        while (Packing.unpackedVertices.size() > t) {
            Pack pack = Packing.addPackGreedy();
            if (pack.size != t + 1) break;
            if (!useWeightedVersion)
                for (int i = 0; i < pack.size; i++) {
                    Vertex vertex = pack.packedVertices[i];
                    if (vertex.inManyPacks != vertex.weight)
                        Packing.unpackedVertices.add(vertex);
                }

            if (pack.cost == Integer.MAX_VALUE) return k + 1;
            lb += pack.cost;
            if ((lb + 1) / 2 > k) break;
        }

        //improve locally
        while ((increase = Packing.improvePackingBySwaps()) != 0) {
            if (flagUnfeasible) {
                flagUnfeasible = false;
                return k + 1;
            }
            lb += increase;
        }
        return (lb + 1) / 2;
    }
}
