package algo.datastructure;

import algo.branch.Packing;

import java.util.Objects;

public class Edge {

    public final Vertex u, v;
    final int hash;

    public boolean isEdge = false;
    public boolean isEdited = false;
    public boolean isBlocked = false;
    //in the code the witnesses are named as non commons since they are not a common neighbor or a vertex pair
    public PartSetVertices nonCommons;

    //merging variables
    public boolean isMerged = false;
    public boolean isApart = false;
    public int vertexWeightsOfNonCommons = 0;

    public Edge(Vertex u, Vertex v, Graph graph) {
        this.u = u;
        this.v = v;
        nonCommons = new PartSetVertices(graph);
        hash = Objects.hash(u, v);
    }

    public int getEditCost() {
        return u.weight * v.weight;
    }

    public boolean containsNonCommonNeighbor(Vertex vertex) {
        return nonCommons.contains(vertex);
    }

    public int getNumberOfNonCommons() {
        return nonCommons.size;
    }

    public int getResolveCostOfNonCommons() {
        return (Packing.useWeightedCost ? vertexWeightsOfNonCommons : nonCommons.size ) * Math.min(u.weight, v.weight);
    }

    public void addNonCommon(Vertex vertex) {
        nonCommons.add(vertex);
        vertex.nonCommonForTheseEdges.add(this);
        vertexWeightsOfNonCommons += vertex.weight;
        Packing.updatePackingByAddNonCommon(this);
    }

    public void removeNonCommon(Vertex vertex) {
        nonCommons.remove(vertex);
        vertex.nonCommonForTheseEdges.remove(this);
        vertexWeightsOfNonCommons -= vertex.weight;
        Packing.updatePackingByRemoveNonCommon(this);
    }

    public void flipNonCommon(Vertex vertex) {
        if (nonCommons.contains(vertex)) removeNonCommon(vertex);
        else addNonCommon(vertex);
    }

    public void updateCommonsAndNonCommonsByEdit(Graph g) {
        for (Vertex vertex : g) {
            if (u == vertex || v == vertex) continue;
            g.getEdge(u, vertex).flipNonCommon(v);
            g.getEdge(v, vertex).flipNonCommon(u);
        }
    }

    @Override
    public boolean equals(Object o) {
        return this == o;
    }

    @Override
    public int hashCode() {
        return hash;
    }

    public void block() {
        isBlocked = true;
    }

    public void unblock() {
        isBlocked = false;
    }

    public void setApartTrue() {
        isApart = true;
        Packing.updatePackingBySetApart(this);
    }

    public void setApartFalse() {
        isApart = false;
        Packing.updatePackingBySetApart(this);
    }

    public boolean isNotMergeableWithLabel() {
        Label conA = u.getLabel();
        Label conB = v.getLabel();

        if (conA == Label.ADJACENT && conB == Label.SEPARATE
                || conA == Label.SEPARATE && conB == Label.ADJACENT) {
            return true;
        }

        if (isEdited || isBlocked) {
            Label negatedAdjacencyOfEdge = (isEdge) ? Label.SEPARATE : Label.ADJACENT;
            return conA == negatedAdjacencyOfEdge || conB == negatedAdjacencyOfEdge;
        }
        return false;
    }
}
