package algo.datastructure;

public class NCs {

    public NC[] ncs; //neighborhood classes
    public int ncCount;

    Graph g;

    NCs(Graph g) {
        this.g = g;
        ncCount = 0;
        //create empty neighborhood classes
        ncs = new NC[g.vertexCount];
        for (int i = 0; i < ncs.length; i++) ncs[i] = new NC(g, i);
        //compute neighborhood partition
        for (Vertex vertex : g) {
            boolean newNC = true;
            for (int i = 0; i < ncCount; i++) {
                NC nc = ncs[i];
                Vertex ncVertex = nc.getRepresentative();
                if (g.areSimilar(vertex, ncVertex)) {
                    nc.addVertex(vertex);
                    newNC = false;
                }
            }
            if (newNC) {
                NC nc = ncs[ncCount++];
                nc.addVertex(vertex);
            }
        }
    }

    public void updateNCsByEdit(Edge e) {
        Vertex u = e.u;
        Vertex v = e.v;
        NC newNCA = null;
        NC newNCB = null;
        //find NCs where u and v fit respectively
        for (int i = 0; i < ncCount; i++) {
            NC nc = ncs[i];
            int index = 0;
            Vertex vertex;
            do {//get vertex of NC that is not u or v
                vertex = nc.get(index);
                index++;
            } while (((vertex == u || vertex == v) && index != nc.vertices.size));

            if (vertex == u || vertex == v) continue;
            if (newNCA == null && g.areSimilar(u, vertex)) {
                newNCA = vertex.currentNC;
            }
            if (newNCB == null && g.areSimilar(v, vertex)) {
                newNCB = vertex.currentNC;
            }
        }
        //special case: both fit in no NC
        if (newNCA == null && newNCB == null) {
            if (u.currentNC == v.currentNC) {
                if (u.currentNC.vertices.size() != 2) {
                    if (g.areSimilar(u, v)) {
                        NC newNC = ncs[ncCount++];
                        moveVertexToNC(u, newNC);
                        moveVertexToNC(v, newNC);
                    } else {
                        moveVertexToNC(u, ncs[ncCount++]);
                        moveVertexToNC(v, ncs[ncCount++]);
                    }
                } else {//old NC of u and v has only them as elements
                    if (!g.areSimilar(u, v))
                        moveVertexToNC(u, ncs[ncCount++]);
                }
                return;
            } else {
                if (g.areSimilar(u, v)) {
                    if (u.currentNC.vertices.size() == 1) {
                        moveVertexToNC(v, u.currentNC);
                        return;
                    }
                    if (v.currentNC.vertices.size() == 1) {
                        moveVertexToNC(u, v.currentNC);
                        return;
                    }
                }
            }
        }
        //move u and v in fitting nc or in new one
        if (u.currentNC.size != 1) {
            if (newNCA == null) {
                moveVertexToNC(u, ncs[ncCount++]);
            } else moveVertexToNC(u, newNCA);
        } else if (newNCA != null) moveVertexToNC(u, newNCA);

        if (v.currentNC.size != 1) {
            if (newNCB == null) {
                moveVertexToNC(v, ncs[ncCount++]);
            } else moveVertexToNC(v, newNCB);
        } else if (newNCB != null) moveVertexToNC(v, newNCB);

    }

    public void searchNCAndMoveVertex(Vertex vertex) {
        if (vertex.currentNC == null) {
            //find fitting NC
            for (int i = 0; i < ncCount; i++) {
                NC nc = ncs[i];
                Vertex ncVertex = nc.getRepresentative();
                if (ncVertex == vertex) continue;
                if (g.areSimilar(vertex, ncVertex)) {
                    ncVertex.currentNC.addVertex(vertex);
                    return;
                }
            }
            ncs[ncCount++].addVertex(vertex);
        } else {
            //there are two cases where no change is done
            for (Vertex mergedVertex : vertex.currentNC.vertices) {//first case: currentNC fits for vertex
                if (vertex != mergedVertex) {
                    if (g.areSimilar(mergedVertex, vertex)) return;
                    break;
                }
            }

            for (int i = 0; i < ncCount; i++) {
                NC nc = ncs[i];
                Vertex ncVertex = nc.getRepresentative();
                if (ncVertex == vertex) continue;
                if (g.areSimilar(vertex, ncVertex)) {
                    moveVertexToNC(vertex, ncVertex.currentNC);
                    return;
                }
            }
            if (vertex.currentNC.vertices.size() == 1) return;//second case: single sized NC
            //no NC fits vertex
            moveVertexToNC(vertex, ncs[ncCount++]);
        }
    }

    public void moveVertexToNC(Vertex vertex, NC destinationNC) {
        NC originNC = vertex.currentNC;
        originNC.removeVertex(vertex);//first remove then add!
        destinationNC.addVertex(vertex);
        updateOfEmptyNCByRemove(originNC);
    }

    public void updateOfEmptyNCByRemove(NC originNC) {
        if (originNC.vertices.size() == 0) {
            ncCount--;
            NC changeNC = ncs[ncCount];
            ncs[originNC.indexForNCArray] = changeNC;
            ncs[ncCount] = originNC;
            changeNC.indexForNCArray = originNC.indexForNCArray;
            originNC.indexForNCArray = ncCount;
        }
    }
}
