package algo.datastructure;

import algo.branch.Pack;
import algo.util.InitialVertex;

import java.util.*;

import static algo.util.Util.over2;

public class Vertex implements Iterable<Vertex> {
    public final int id;
    public final String name;
    private final int hash;

    public final PartSetVertices neighbors;
    public Block currentBlock;

    public NC currentNC = null;
    public NC initialNC = null;

    public int weight = 1;
    public Label label = Label.UNDECIDED;

    public Set<Edge> nonCommonForTheseEdges = new HashSet<>();


    //block index
    public int indexOfBlockArrayPosition = -1;

    //data structure for packs
    public int inManyPacks = 0;
    public int[] indicesInPacks = new int[10];
    public Pack[] packs = new Pack[10];

    public int insideEdges = 0;
    public int apartEdgesCount = 0;

    Stack<Vertex> parentVertices = new Stack<>();

    public Vertex(InitialVertex initialVertex, int maxNumberOfVertices) {
        this.id = initialVertex.id;
        this.name = initialVertex.name;
        hash = Objects.hash(id);
        neighbors = new PartSetVertices(maxNumberOfVertices);
        Arrays.fill(indicesInPacks, -1);
    }

    public int degree() {
        return neighbors.size;
    }

    @Override
    public boolean equals(Object o) {
        return this == o;
    }

    @Override
    public int hashCode() {
        return hash;
    }

    public int getIndexOfNC() {
        return currentNC.indexForNCCostArray;
    }

    public void setParentVertex(Vertex parent) {
        parentVertices.push(parent);
    }

    public void removeLastParentVertex() {
        parentVertices.pop();
    }

    public Vertex getParentVertex() {
        if (parentVertices.isEmpty()) return this;
        else return parentVertices.peek().getParentVertex();
    }

    public Label getInsideAdjacency() {
        Label con = getLabel();
        if (con != Label.UNDECIDED) return con;
        //order of cases is important
        if (weight == 1) return Label.UNDECIDED;
        if (insideEdges == 0) return Label.SEPARATE;
        if (insideEdges == over2(weight)) return Label.ADJACENT;
        return Label.UNDECIDED;
    }

    public Label getLabel() {
        return label;
    }

    public int computeInsideCost() {
        Label con = getInsideAdjacency();
        switch (con) {
            case ADJACENT -> {
                return over2(weight) - insideEdges;
            }
            case SEPARATE -> {
                return insideEdges;
            }
            default -> {// case UNDECIDED
                return Math.min(insideEdges, over2(weight) - insideEdges);
            }
        }
    }

    public int getBlockIndex() {
        return currentBlock.index;
    }

    @Override
    public Iterator<Vertex> iterator() {
        return neighbors.iterator();
    }

    //code for pack data structure
    public void swapPackWithLastPack(int i) {
        int lastEntry = inManyPacks - 1;

        Pack auxPack = packs[i];
        int auxInt = indicesInPacks[i];

        packs[i] = packs[lastEntry];
        indicesInPacks[i] = indicesInPacks[lastEntry];

        packs[lastEntry] = auxPack;
        indicesInPacks[lastEntry] = auxInt;

        packs[i].indicesOfPacksArrayOfVertex[indicesInPacks[i]] = i;

        packs[lastEntry]
                .indicesOfPacksArrayOfVertex[
                indicesInPacks[lastEntry]] = lastEntry;
    }

    public void removeLastPack() {
        inManyPacks--;
        indicesInPacks[inManyPacks] = -1;
        packs[inManyPacks] = null;
    }

    public void expandBlockArrayIfNecessary() {
        if (inManyPacks == packs.length) {
            Pack[] oldPacks = packs;
            packs = new Pack[(int) (inManyPacks * 1.5)];
            System.arraycopy(oldPacks, 0, packs, 0, inManyPacks);

            int[] oldIndices = indicesInPacks;
            indicesInPacks = new int[(int) (inManyPacks * 1.5)];
            System.arraycopy(oldIndices, 0, indicesInPacks, 0, inManyPacks);
        }
    }

}
