import networkx as nx
import numpy as np

# https://networkx.org/documentation/stable/reference/classes/index.html
non_iso_matrices = list()
perms = list()
ctr = 0

def test_graph(g):
    global ctr
    ctr += 1
    if exists_same_neighborhood(g):
        return False
    mat = create_adj_matrix(g)
    if not is_isomorph(mat):
        non_iso_matrices.append(mat)
        print(ctr)


def create_adj_matrix(g):
    mat = np.full([len(g), len(g)], 0)
    for e in g.edges():
        mat[e[0]][e[1]] = 1
        mat[e[1]][e[0]] = 1
    return mat


def set_permutations_variable(size):
    global perms
    perms = list()
    with open(f"permutations_{size}.txt", 'r') as file:
        for line in file:
            splitted = line.split()
            arr = np.array(list(map(lambda x: int(x), splitted)))
            perms.append(arr)
    return perms


def is_isomorph(mat):
    is_candidate = True
    for matB in non_iso_matrices:
        if is_isomorph_with(mat, matB):
            is_candidate = False
            break
    return not is_candidate


def is_isomorph_with(matA, matB):
    for perm in perms:
        fit = True
        for i in range(len(matA)):
            if not fit: break
            for j in range(i, len(matA)):
                if matA[perm[i]][perm[j]] != matB[i][j]:
                    fit = False
                    break
        if fit: return True
    return False


def create_graphs(graph, index):
    if index == len(graph):
        test_graph(graph)
        return
    graph.add_edge(index, index)
    create_graphs(graph, index + 1)
    graph.remove_edge(index, index)
    create_graphs(graph, index + 1)


def exists_same_neighborhood(g):
    for i in range(len(g)):
        for j in range(i + 1, len(g)):
            if have_same_neighborhood(g, i, j):
                return True


def have_same_neighborhood(g, u, v):
    if g.degree[u] != g.degree[v]:
        return False
    selfloopU = g.has_edge(u, u)
    selfloopV = g.has_edge(v, v)
    if selfloopV != selfloopU:
        return False
    if selfloopU != g.has_edge(u, v):
        return False
    addCounter = 0
    if selfloopU:
        addCounter += 3 ##### NOTE: a selfloop increses the degree of a node by 2 and edge {u, v} is not count in nx.common_neighbors########

    common_neighbors = sum(1 for dummy in nx.common_neighbors(g, u, v))

    return common_neighbors + addCounter == g.degree[u]


def write_non_iso_matrices():
    with open(f"non_iso_mats_{len(non_iso_matrices[0])}.txt", 'w') as file:
        ctr = 1
        for mat in non_iso_matrices:
            file.write(f"# {ctr}\n")
            ctr += 1
            for row in mat:
                res = list(map(lambda x: str(x), row))
                file.write(f"{' '.join(res)}\n")



if __name__ == '__main__':
    # get all graphs with size
    size = 6
    l = nx.graph_atlas_g()
    fl = list(filter(lambda graph: len(graph.nodes) == size, l))
    set_permutations_variable(size)

    for g in fl:
        create_graphs(g, 0)
    write_non_iso_matrices()
