package algo.util;

import algo.datastructure.*;
import algo.heuristic.BlockPartitionResult;
import algo.heuristic.Heuristic;

import java.lang.reflect.Array;
import java.util.*;

public class UtilTesting {

    public static void printGraphViz(Graph g) {
        for (Vertex v : g) {
            for (Vertex neighbor : v) {
                if (neighbor.id > v.id) {
//                    System.out.println(String.format("%s -- %s", v.id, neighbor.id));
                    System.out.println(String.format("%s -- %s", v.name, neighbor.name));

                }
            }
            if (v.degree() == 0)
                System.out.println(v.name);
        }
    }

    public static String printColor(Graph g, int t) {
        //für karate
        int[] group1 = new int[]{1, 2, 3, 4, 5, 6, 7, 8, 9, 11, 12, 13, 14, 17, 18, 20, 22};
        int[] group2 = new int[]{10, 15, 16, 19, 21, 23, 24, 25, 26, 27, 28, 29, 30, 31, 32, 33, 34};

        List<String> gr1 = new ArrayList<>();
        List<String> gr2 = new ArrayList<>();

        for (int i : group1) gr1.add(Integer.toString(i));
        for (int i : group2) gr2.add(Integer.toString(i));

//        System.out.println(gr1);
//        if (true) return "";

        StringBuilder res = new StringBuilder();
        res.append("graph G {\n");
        for (Vertex v : g) {
            for (Vertex neighbor : v) {
                if (neighbor.id > v.id)
                    res.append(String.format("%s -- %s\n", v.name, neighbor.name));
            }
            if (v.degree() == 0)
                res.append(v.name).append("\n");
        }

        BlockPartitionResult result = Heuristic.computeUpperBoundWithBlockPartition(g, t);
        System.out.println("upper bound is " + result.getCost());
        BlockPartition bp = result.getBlockPartition();
        Character[][] bm = new Character[bp.size()][bp.size()];
        String[] colors = new String[]{"red", "cyan", "green","orange", "yellow", "gray", "blue" , "purple"};
        List<Block> blocks = bp.getBlocks();
        for (int b1Index = 0; b1Index < blocks.size(); b1Index++) {
            Block b1 = blocks.get(b1Index);
            String col = colors[b1Index];
            Vertex[] verticesArray = b1.verticesArray;
            for (int vertexIndex = 0; vertexIndex < b1.size(); vertexIndex++) {
                Vertex vertex = verticesArray[vertexIndex];
                String shape = "";
                if (gr1.contains(vertex.name)){
                    shape = ", shape=box";
                }
                res.append(String.format("%s [color=\"%s\", style=filled%s]\n", vertex.name, col, shape));

            }

            List<Block> bpBlocks = bp.getBlocks();
            for (int b2Index = 0; b2Index < bpBlocks.size(); b2Index++) {
                Block b2 = bpBlocks.get(b2Index);
                int edges = bp.getEdgeCountBetween(b1, b2);
                double maxEdges;
                if (b1 == b2) maxEdges = Util.over2(b1.size());
                else maxEdges = b1.size() * b2.size();
                char con = ' ';
                if (edges > maxEdges / 2) {
                    con = '1';
                }
                if (edges < maxEdges / 2) {
                    con = '0';
                }
                if (edges == maxEdges / 2) {
                    con = '?';
                }
                bm[b1Index][b2Index] = con;
            }
        }
        if (true){
            res.append("subgraph cluster_1 {\n");
            for (int b1Index = 0; b1Index < blocks.size(); b1Index++) {
                String col1 = colors[b1Index];
                res.append(String.format("%s [color=\"%s\", style=filled]\n", "Block" + b1Index, col1));
            }

            for (int b1Index = 0; b1Index < blocks.size(); b1Index++) {
                List<Block> bpBlocks = bp.getBlocks();
                for (int b2Index = b1Index; b2Index < bpBlocks.size(); b2Index++) {
                    char con = bm[b1Index][b2Index];
                    if (con == '1')
                        res.append(String.format("%s -- %s\n", "Block" + b1Index, "Block" + b2Index));
                    if (con == '?')
                        res.append(String.format("%s -- %s [style=dashed]\n", "Block" + b1Index, "Block" + b2Index));
                }
            }
            res.append("label = \"Block Model Graph\";\n" + "}");
        }
        res.append("}");
        return res.toString();
    }

    public static void printAdjacencyMatrix(Graph g, boolean sortByNcs) {
        Vertex[] vertices = new Vertex[g.vertexCount];
        int index = 0;
        Set<Integer> markerForHLine = new HashSet<>();
        if (sortByNcs) {
            for (int i = 0; i < g.ncs.ncCount; i++) {
                NC nc = g.getNC(i);

                List<Vertex> verticesOfNC = new ArrayList<>(nc.vertices);
                verticesOfNC.sort(Comparator.comparingInt(x -> x.id));
                for (Vertex vertex : verticesOfNC) {
                    vertices[index] = vertex;
                    index++;
                }
//                for (Vertex vertex : nc.vertices) {
//                    vertices[index] = vertex;
//                    index++;
//                }

                markerForHLine.add(index);
            }
        } else {
            for (Vertex vertex : g) {
                vertices[index] = vertex;
                index++;
            }
        }

        //start line
        StringBuilder startLine = new StringBuilder();
        startLine.append("  & ");
        for (Vertex u : vertices) {
            startLine.append(u.id).append(" & ");
        }
        startLine.deleteCharAt(startLine.length() - 1);
        startLine.deleteCharAt(startLine.length() - 1);
        startLine.append("\\\\ \\hline");
        System.out.println(startLine);//new line
        int indexHLine = 0;
        for (Vertex u : vertices) {
            indexHLine++;

            StringBuilder line = new StringBuilder();
            line.append(u.id).append(" & ");

            for (Vertex v : vertices) {
                if (u == v) {
                    line.append("-");
                } else {
                    if (g.areInitialNeighbors(u, v))
                        line.append("1");
                    else line.append("0");
                }
                line.append(" & ");
            }
            line.deleteCharAt(line.length() - 1);
            line.deleteCharAt(line.length() - 1);

            line.append("\\\\");
            if (markerForHLine.contains(indexHLine))
                line.append(" \\hline");
            System.out.println(line);//new line

        }


        if (sortByNcs) {//current matrix
            for (Vertex u : vertices) {
                for (Vertex v : vertices) {
                    if (u == v) {
                        System.out.print("-");
                        continue;
                    }
                    if (g.areNeighbors(u, v))
                        System.out.print("1");
                    else
                        System.out.print("0");
                }
                System.out.println();//new line
            }
        }
    }
}
